package com.grupogodo.corp.projects.optimus.composers;

import java.util.*;
import java.util.Calendar;

import javax.annotation.PostConstruct;

import lombok.Getter;
import lombok.val;
import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.Executions;
import org.zkoss.zk.ui.event.*;
import org.zkoss.zk.ui.event.EventListener;
import org.zkoss.zk.ui.select.annotation.Listen;
import org.zkoss.zk.ui.select.annotation.Wire;
import org.zkoss.zul.*;
import org.zkoss.zul.event.TreeDataEvent;

import com.grupogodo.corp.libs.jpa.core.BaseEntity;
import com.grupogodo.corp.libs.springzk.scope.DesktopScoped;
import com.grupogodo.corp.libs.springzk.spring.Reload;
import com.grupogodo.corp.libs.springzk.spring.ZKModelAwareComposer;
import com.grupogodo.corp.libs.zkcomponents.components.growl.Growl;
import com.grupogodo.corp.libs.zkcomponents.components.growl.Growl.Severity;
import com.grupogodo.corp.projects.optimus.entities.*;
import com.grupogodo.corp.projects.optimus.entities.Group;
import com.grupogodo.corp.projects.optimus.enums.ExecutionResult;
import com.grupogodo.corp.projects.optimus.services.*;
import com.grupogodo.corp.projects.optimus.services.processor.ProcessorExecutor;

@Slf4j
@DesktopScoped
public class ExecutionComposer extends ZKModelAwareComposer<Component> {

    private static final long serialVersionUID = 2345427318969691259L;

    private class ProcessTreeModel extends DefaultTreeModel<BaseEntity> {

        private static final long serialVersionUID = -6919770287544593521L;

        public ProcessTreeModel(TreeNode<BaseEntity> root) {
            super(root);
        }

        public void invalidate() {
            fireEvent(TreeDataEvent.STRUCTURE_CHANGED, getPath(ghostTreeNode), 0, 0);
        }
    }

    private class ProcessItemRenderer implements TreeitemRenderer<TreeNode<BaseEntity>> {

        @Override
        public void render(Treeitem item, TreeNode<BaseEntity> node, int index) throws Exception {
            Treerow treeRow = new Treerow();
            Treecell cell = new Treecell();
            Label label = new Label();
            cell.appendChild(label);
            treeRow.appendChild(cell);

            if (node.getData() instanceof Group) {
                Group group = (Group) node.getData();
                label.setValue(group.getName());
                item.setContext("groupContext");
                cell.setImage("/resources/images/menu_tree/group.png");

            } else if (node.getData() instanceof ProcessDefinition) {
                ProcessDefinition processDefinition = (ProcessDefinition) node.getData();
                label.setValue(processDefinition.getDescription());
                item.setContext("processContext");
                cell.setImage("/resources/images/menu_tree/process.png");

            }

            Boolean openStatus = treeOpenStatusMap.get(node.getData().getId());
            item.setOpen(openStatus == null ? false : openStatus);

            item.setAttribute("id", node.getData().getId());
            label.setStyle("margin: 2px");
            item.setValue(node);
            item.appendChild(treeRow);

            TreeEventListener treeEventListener = new TreeEventListener();
            item.addEventListener(Events.ON_CLICK, treeEventListener);
            item.addEventListener(Events.ON_DOUBLE_CLICK, treeEventListener);
            item.addEventListener(Events.ON_RIGHT_CLICK, treeEventListener);
        }
    }

    private class TreeEventListener implements EventListener<Event> {

        @Override
        public void onEvent(Event event) throws Exception {
            Treeitem currentItem = (Treeitem) event.getTarget();
            currentNode = currentItem.getValue();

            Include include = new Include("/WEB-INF/views/executionGroupForm.zul");
            include.setDynamicProperty("data", currentNode.getData());
            detailRegion.getChildren().clear();
            include.setParent(detailRegion);
            
            // Mostrar los procesos que contiene al clickar en grupos
            if (currentNode.getData() instanceof Group) {

                processExecuteButton.setDisabled(true);
                currentGroup = (Group) currentNode.getData();
                executionModel.clear();

                executionModel.addAll(executionService.findByGroup(currentGroup));

                // Mostrar las ejecuciones de los procesos al clickar en procesos
            } else if (currentNode.getData() instanceof ProcessDefinition) {

                processExecuteButton.setDisabled(false);
                currentProcess = (ProcessDefinition) currentNode.getData();
                executionModel.clear();

                executionModel.addAll(executionService.findByProcessDefinition(currentProcess));
            }
        }
    }

    @Autowired
    private ProcessService processService;
    
    @Autowired
    private ProcessorExecutor processorExecutor;

    @Autowired
    private GroupService groupService;

    @Autowired
    private ExecutionService executionService;
    
    @Wire
    private Toolbarbutton processExecuteButton;

    @Getter
    private ProcessItemRenderer itemRenderer;

    @Getter
    private ProcessTreeModel processTreeModel;

    @Getter
    private ListModelList<ProcessDefinition> processModel;

    @Getter
    private ListModelList<Execution> executionModel;

    @Wire
    private Div detailRegion;

    private Map<Long, Boolean> treeOpenStatusMap = new HashMap<Long, Boolean>();

    private TreeNode<BaseEntity> ghostTreeNode;

    private TreeNode<BaseEntity> currentNode;

    private Group currentGroup;

    private ProcessDefinition currentProcess;

    @PostConstruct
    @Reload
    public void init() {
        processModel = new ListModelList<ProcessDefinition>();
        executionModel = new ListModelList<Execution>();
        buildProcessModel();
    }

    @Listen("onClick = #processExecuteButton; onClick = #processExecuteContext; onClick = #newExecutionButton")
    public void executeProcess() {

        if (currentNode.getData() instanceof Group) {
            //cambiar esto por un mensaje de log / warning
            log.warn("No es pot executar un Grup");
            
        } else {
            ProcessDefinition processDefinition = ((ProcessDefinition)currentNode.getData());
            Execution execution = new Execution(processDefinition, Calendar.getInstance().getTime(), ExecutionResult.KO);
            try {
                processorExecutor.execute(processDefinition.getName());
                execution.setExecutionResult(ExecutionResult.OK);
                Growl.showMessageI18n("ui.message.executionOK", "ui.message.executionOK.description", 2000, Severity.INFO) ;
                
            } catch (Exception e) {
                Growl.showMessageI18n("ui.message.errorExecution", "ui.message.errorExecution.description", 2000, Severity.WARN);
                
                //si falla algo, añadimos la excepción
                ExecutionException executionException = new ExecutionException(e.getClass().toString(), e.getMessage(), ExceptionUtils.getStackTrace(e), execution, processDefinition);
                execution.setExecutionException(executionException);
            } finally {
                executionService.save(execution);
                executionModel.clear();
                executionModel.addAll(executionService.findByProcessDefinition(currentProcess));
            }
        }
    }

    @Listen("onClick = #expandAllButton")
    public void expandTree() {
        expandTreeRecursively(ghostTreeNode);
    }
    
    
    @Listen("onGoToProcessDefinition = #executionListBox")
    public void onGoToProcessDefinition(Execution execution) {
        Executions.sendRedirect("/process/" + execution.getProcessDefinition().getName());
    }
    
    public boolean isProcessDefinition(BaseEntity node) {
        return node instanceof ProcessDefinition ? true : false;
    }
    
    private void expandTreeRecursively(TreeNode<BaseEntity> tree) {
        //TODO
//        tree.open();
//        for (tree.children) {
//            expandTreeRecursively(child);
//        }
    }

    private void buildProcessModel() {

        val rootGroup = groupService.getRoot();

        ghostTreeNode = new DefaultTreeNode<BaseEntity>(rootGroup, new LinkedList<TreeNode<BaseEntity>>());
        processTreeModel = new ProcessTreeModel(ghostTreeNode);
        itemRenderer = new ProcessItemRenderer();
        buildTreeRecursively(ghostTreeNode, rootGroup);

    }

    // Recursividad de grupos, cada grupo apunta a su parent (grupo padre) y se forma el arbol desde el primer nodo de grupo llamado
    // rootGroup.
    private void buildTreeRecursively(TreeNode<BaseEntity> parentNode, Group parent) {
        List<Group> subgroups = groupService.findByParent(parent);

        for (Group group : subgroups) {
            TreeNode<BaseEntity> node = addElementToTree(parentNode, group);
           val processes = group.getProcesses();
            for (ProcessDefinition processDefinition : processes) {
                // mostramos en el arbol solo los procesos ejecutables
                if (!processDefinition.isDeleted() && (processDefinition.isExecutable())) {
                    addElementToTree(node, processDefinition);
                }
            }
            buildTreeRecursively(node, group);
        }
    }

    private TreeNode<BaseEntity> addElementToTree(TreeNode<BaseEntity> parentNode, BaseEntity entity) {
        TreeNode<BaseEntity> entityNode = null;
        if (entity instanceof ProcessDefinition) {
            entityNode = new DefaultTreeNode<BaseEntity>(entity);
        } else {
            entityNode = new DefaultTreeNode<BaseEntity>(entity, new LinkedList<TreeNode<BaseEntity>>());
        }
        parentNode.add(entityNode);
        processTreeModel.invalidate();

        return entityNode;
    }
}
